# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
from tinyms import layers
from tinyms.primitives import Sigmoid


class DeepFMEvalModel(layers.Layer):
    """
    Provide DeepFM training network.

    Args:
        network (layers.Layer): The base network.

    Returns:
        logits (Tensor) - With the shape as [batch_size, vocab_size] and data type as float32.
        predict_probs (Tensor) - With the same shape as [batch_size, vocab_size] and data type as float32.
        labels (Tensor) - With the same shape as labels.

    Examples:
        >>> from tinyms.model import deepfm, DeepFMEvalModel
        >>>
        >>> net = deepfm()
        >>> eval_net = DeepFMEvalModel(net)
    """
    def __init__(self, network):
        super(DeepFMEvalModel, self).__init__(auto_prefix=False)
        self.network = network
        self.sigmoid = Sigmoid()

    def construct(self, ids, wts, labels):
        logits, _, _ = self.network(ids, wts)
        predict_probs = self.sigmoid(logits)
        return logits, predict_probs, labels
