# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
.. TinyMS package.

Top-level reference to dtype of common module.
This module also provides Numpy-like interfaces in TinyMS.

Examples:
    >>> import tinyms as ts
    >>>
    >>> print(ts.ones([2, 3]))
    [[1. 1. 1.]
    [1. 1. 1.]]
"""
from .version import __version__
from . import common
from .common import *

__all__ = []
__all__.extend(__version__)
__all__.extend(common.__all__)
